"""Incoming Webhooks

This module contains actions for sending messages & approval requests to a Slack channel.

!!! tip "Before you Start"

    Ensure you have created a PPA Slack app, & enabled incoming webhooks.

    See the [Integration](../integration/) page for more information
"""

import requests
import logging
import datetime
import shortuuid

from typing import Dict, Optional, Any, Union

from ppa import appliance, task, events

from . import exceptions
from .records import WaitHookResponse


logger = logging.getLogger(__name__)


def _check_status_code(response) -> requests.Response:
    status_code = response.status_code

    if status_code == 400:
        logger.error(response.text)
        raise exceptions.InvalidMessage("The supplied message format is invalid.")

    elif status_code in {403, 404}:
        logger.error(response.text)
        raise exceptions.InvalidWebhook(
            f"Requested to the supplied webhook URL returned status code '{status_code}' with message '{response.text}'."
        )

    return response


def send_message(url: str, message: Optional[Union[Dict[str, Any], str]] = None) -> None:
    """Send a message to a Slack channel.

!!! question "Slack Channel & Rich Messages"

    The message will be send to the channel associated with the supplied webhook URL.

    View the Slack block kit builder [here](https://app.slack.com/block-kit-builder/) for more information on creating rich messages.

??? info "Input"

    - __url:__ the Slack incoming webhook URL

    - __message:__ the message to send to the channel

??? success "Output"

    Nothing is outputted by this action.

??? example "Simple Message"

    ```YAML linenums="1"
    - slack.incoming_webhooks.send_message:
        message: Hello from PPA!
      load:
        url: slack_secrets.webhook_url
    ```

??? example "Markdown Message"

    ```YAML linenums="1"
    - slack.incoming_webhooks.send_message:
        message: blocks:
        - type: section
          text:
            type: mrkdwn
            text: 'Hello from PPA!'
      load:
        url: slack_secrets.webhook_url

    ```

!!! warning "Supplying Sensitive Data"

    You should ___always___ use a PPA Vault integration to provide sensitive data to a plugin action.
    """
    if isinstance(message, str):
        message = {"text": message}

    _check_status_code(requests.post(url, json=message))


def send_approval_request(url: str, message: str) -> WaitHookResponse:
    """Send an approval request to a Slack channel & wait until it approved/rejected.

Once the request is responded to, a message will be posted to Slack containing the following:

- Whether the request was approved or rejected
- The username of the person who responded to the request

Example screenshots of all the Slack messages can be found under the example.

!!! question "Slack Channel"

    The message will be send to the channel associated with the supplied webhook URL.

??? info "Input"

    - __url:__ the Slack incoming webhook URL

    - __message:__ a message to include in the approval request

??? success "Output"

    A dictionary with the following keys:

    - __approved:__ `true` if the request was approved, `false` if it was not

    - __responder:__ the PPA user who submitted the response

??? example

    ```YAML linenums="1"
    - slack.incoming_webhooks.send_approval_request:
        message: Add user to Remote Desktop Users group.
      load:
        url: slack_secrets.webhook_url
      save: approval_response
    ```

??? note "Approval Request Screenshot"

    <img src="../../../images/screenshots/slack/approval_request.png" alt="approval-request" width="100%"/>

??? note "Approved Message Screenshot"

    <img src="../../../images/screenshots/slack/request_approved.png" alt="requested-approved" width="100%"/>

??? note "Rejected Message Screenshot"

    <img src="../../../images/screenshots/slack/request_rejected.png" alt="request-rejected" width="100%"/>

!!! warning "Supplying Sensitive Data"

    You should ___always___ use a PPA Vault integration to provide sensitive data to a plugin action.
    """
    wait_hook_url, secret = events.create_wait_hook()
    approve_link = f"{wait_hook_url}?approve=true"
    reject_link = f"{wait_hook_url}?approve=false"
    task_link = f"https://{appliance.address()}/ui/activity/attach/?uuid={wait_hook_url.split('/')[-2]}"
    task_name = task.manifest()["labels"]["ppa.name"]
    request_uuid = shortuuid.uuid()

    try:
        started_by = task.user().split("\\")[1]
    except IndexError:
        started_by = task.user()

    request_info = f"""*Request Info*

Task: {task_name}

Operator: {started_by}

Requested At: {datetime.datetime.utcnow().strftime('%H:%M:%S %d/%m/%Y')}

Unique Request ID: {request_uuid}

Click <{task_link}|here> to view the task.
    """

    request_message = f"""*Request Message*

{message}
"""

    request_actions = f"""*Please Choose an Option*


:white_check_mark: <{approve_link}|Approve>


:x: <{reject_link}|Reject>"""

    request_body = {
        "blocks": [
            {"type": "divider"},
            {"type": "section", "text": {"type": "mrkdwn", "text": ":bell: *New PPA Approval Request* :bell:"}},
            {"type": "divider"},
            {"type": "section", "text": {"type": "mrkdwn", "text": request_info}},
            {"type": "divider"},
            {"type": "section", "text": {"type": "mrkdwn", "text": request_message}},
            {"type": "divider"},
            {"type": "section", "text": {"type": "mrkdwn", "text": request_actions}},
            {
                "type": "section",
                "text": {"type": "mrkdwn", "text": "_You will be asked to authenticate to PPA before your choice is submitted._"},
            },
            {"type": "divider"},
        ]
    }

    _check_status_code(requests.post(url, json=request_body))

    approval_response = WaitHookResponse(**events.wait("Waiting for Slack approval response...", secret=secret))
    approved = approval_response["values"]["approve"][0].lower() == "true"

    try:
        responder = approval_response["username"].split("\\")[1]
    except IndexError:
        responder = approval_response["username"]

    if not approved:
        approval_response_message = f""":rotating_light: *Request Rejected* :rotating_light:

Rejected By: {responder}

Task: {task_name}

Operator: {started_by}

Unique Request ID: {request_uuid}"""

    else:
        approval_response_message = f""":trophy: *Request Approved* :trophy:

Approved By: {responder}

Task: {task_name}

Operator: {started_by}

Unique Request ID: {request_uuid}"""

    send_message(
        url,
        {
            "blocks": [
                {"type": "divider"},
                {"type": "section", "text": {"type": "mrkdwn", "text": approval_response_message}},
                {"type": "divider"},
            ]
        },
    )

    return {"approved": approved, "responder": responder}
